from wfs20 import __path__
from wfs20._dbase import execute_read_query
from wfs20.error import WFSError

import sqlite3
from pathlib import Path

def _OrderFromDB(code):
	"""
	Get the order out of the database by EPSG code
	"""

	conn = sqlite3.connect(Path(__path__[0],"data","axisorder.db"))
	search_query = f"""\
SELECT * FROM axisorder WHERE code = '{code}';
"""
	r = execute_read_query(conn, search_query)
	conn.close()
	return r[0]

class CRS:
	"""CRS object

	Parameters
	----------
	crs : str
		crs in urn format or uri format
	
	Returns
	-------
	CRS object
	"""

	def __init__(self,crs):
		self.crs = crs
		self.na = "ogc"
		if "urn:" in self.crs:
			self.encoding = "urn"
			s = self.crs.split(":")
			self.na = s[1]
			self.auth = s[-3]
			self.version = s[-2]
			self.code = s[-1]
		elif "/def/crs/" in self.crs:
			self.encoding = "uri"
			s = self.crs.split("/")
			self.auth = s[-3].upper()
			self.version = s[-2]
			self.code = s[-1]
		elif "#" in self.crs:
			self.encoding = "uri"
			s = self.crs.split("/")
			self.auth = s[-1].split(".")[0].upper()
			self.code = s[-1].split("#")[-1]

		try:
			self.order = _OrderFromDB(self.code)[2]
		except IndexError:
			self.order = "xy"

	def __repr__(self):
		return f"<wfs20.crs.CRS object ({self.auth}:{self.code})>"

	def __eq__(self, other):
		if isinstance(self,other.__class__):
			return self.GetURNCode() == other.GetURNCode()
		else:
			return False

	@classmethod
	def from_epsg(cls,code,version=None):
		"""CRS object

		Parameters
		----------
		code : str
			Projection code according to EPSG
		"""
		
		if not version:
			version = ""
		crs_string = f"urn:ogc:def:crs:EPSG:{version}:{code}"
		return cls(crs_string)

	def GetURNCode(self):
		if self.version == 0:
			self.version = ""
		return f"urn:{self.na}:def:crs:{self.auth}:{self.version}:{self.code}"

	def GetURICode1(self):
		if not self.version:
			self.version = 0
		return f"http://www.opengis.net/def/crs/{self.auth}/{self.version}/{self.code}"
	


from wfs20.error import WFSError
from wfs20.util import _PostElement, WFS_NAMESPACE

import sys
import requests
from lxml import etree
from urllib.parse import parse_qsl, urlencode

def _BaseRequestURL(url):
	"""Separate the url in a base-url and parameters
	"""
	
	par = []
	if url.find("?") != -1:
		par = parse_qsl(url.split("?")[1])
	return url.split("?")[0],par

def _ServiceURL(url, version):
	"""Ensure a suitable url for the service GetCapabilities request
	"""

	base,par = _BaseRequestURL(url)
	key = [item[0] for item in par]
	if "service" not in key:
		par += [("service","WFS")]
	if "request" not in key:
		par += [("request","GetCapabilities")]
	if "version" not in key:
		par += [("version",version)]
	urlpar = urlencode(par)
	return "?".join([url.split("?")[0],urlpar])

def GetResponse(
	url: str,
	timeout: int,
	method: str ="GET",
	data: str=None,
) -> requests.models.Response:
	"""Get the response from a url to be requested

	Parameters
	----------
	url : str
		url to be requested
	timeout : int
		Allowed timeout after which an Exception is raised
	method : str, optional
		Request method, either 'GET' or 'POST'
	data : str, optional
		Parameters in xml format

	Returns
	-------
	requests.models.Response
		object holding the response data
	"""

	params = {}
	params["timeout"] = timeout

	if data is not None:
		params["data"] = data

	r = requests.request(method,url,**params)

	if r.status_code in range(400,451,1):
		raise WFSError("Client Error", r.status_code, r.text)
	elif r.status_code in range(500,511,1):
		raise WFSError("Server Error", r.status_code, r.text)

	if "Content-Type" in r.headers and \
			r.headers['Content-Type'] in ['text/xml', 'application/xml', 'application/vnd.ogc.se_xml']:
		wfse = etree.fromstring(r.content)
		exceptions = [
            '{http://www.opengis.net/ows}Exception',
            '{http://www.opengis.net/ows/1.1}Exception',
            '{http://www.opengis.net/ogc}ServiceException',
            'ServiceException'
        ]

		if any(map(wfse.find,exceptions)):
			raise WFSError("WFS Error", r.status_code, r.text)

	# sys.stdout.write(f"Status Code: {r.status_code}\n")

	return r

def BBOXGet(
	bbox: tuple,
	crs: int,
) -> str:
	"""Create string from bounding box tuple
	"""

	if crs.encoding == "urn":
		if crs.order == "xy":
			return "{},{},{},{},{}".format(
				*bbox,
				crs.GetURNCode()
				)
		else:
			return "{},{},{},{},{}".format(
				bbox[1],
				bbox[0],
				bbox[3],
				bbox[2],
				crs.GetURNCode()
				)
	else:
		return "{},{},{},{},{}".format(
			*bbox,
			crs.GetURICode1()
			)

def CreateGetRequest(
	url: str,
	version: str,
	featuretype: str,
	bbox: tuple,
	crs: 'wfs20.crs.CRS',
	startindex: int=None,
) -> str:
	"""Create a geospatial data get request-url

	Parameters
	----------
	url : str
		Service url
	version : str
		Service version
	featuretype : str
		Layer to be requested, mostly in the format of 'xxx:xxx'
	bbox : tuple
		Bounding box wherein the spatial data lies that is requested,
		e.g. (x1,y1,x2,y2)
	crs : wfs20.crs.CRS
		Object containing projection information
	startindex : int, optional
		Starting index of the feature count

	Returns
	-------
	str
		get request url
	"""

	base,_ = _BaseRequestURL(url)
	params = {
		"service":"WFS","version":f"{version}",
		"request":"GetFeature"
		}
	params["typenames"] = [featuretype] 
	params["bbox"] = BBOXGet(bbox, crs)
	if startindex is not None:
		params["startindex"] = startindex
	p = urlencode(params,doseq=True)
	return f"{base}?{p}"

# ToDo: Fix post requests for this library
def CreatePostRequest(
	url: str,
	version: str,
	featuretype: str,
	bbox: tuple,
	crs: 'wfs20.crs.CRS',
	startindex: int=None,
) -> tuple:
	"""Generate post request-url & data

	Parameters
	----------
	url : str
		Service url
	version : str
		Service version
	featuretype : str
		Layer to be requested, mostly in the format of 'xxx:xxx'
	bbox : tuple
		Bounding box wherein the spatial data lies that is requested,
		e.g. (x1,y1,x2,y2)
	crs : wfs20.crs.CRS
		Object containing projection information
	startindex : int, optional
		Starting index of the feature count

	Returns
	-------
	tuple
		Containing base url and the data in XML format
	"""

	base, _ = _BaseRequestURL(url)
	elem = _PostElement(WFS_NAMESPACE, "GetFeature")
	# set the data
	elem.FeatureType(featuretype)
	elem.BBOXPost(bbox, crs)
	if startindex is not None:
		elem.StartIndex(startindex)

	return base, elem.ToString()



class WFSError(Exception):
	def __init__(self,base,code,message):
		self.base = base
		self.code = code
		self.message = message
		super(WFSError,self).__init__(self.base)
	def __str__(self):
		return f"{self.base} => {self.message} (status_code: {self.code})"

class WFSInternalError(Exception):
	def __init__(self,base,message):
		self.base = base
		self.message = message
		super(WFSInternalError, self).__init__(self.base)
	def __str__(self):
		return f"{self.base} => {self.message}"

class AUTHError(Exception):
	def __init__(self,msg):
		self.base = "Wrong authentication credentials."
		self.msg = msg
		super(AUTHError,self).__init__(self.base)
	def __str__(self):
		return f"{self.base} => {self.msg}"


from wfs20.crs import CRS

from collections import defaultdict
from lxml import etree

WFS_NAMESPACE = 'http://www.opengis.net/wfs/2.0'
OWS_NAMESPACE = 'http://www.opengis.net/ows/1.1'
OGC_NAMESPACE = 'http://www.opengis.net/ogc'
GML_NAMESPACE = 'http://www.opengis.net/gml/3.2'
FES_NAMESPACE = 'http://www.opengis.net/fes/2.0'
XSI_NAMESPACE = 'http://www.w3.org/2001/XMLSchema-instance'
XLI_NAMESPACE = 'http://www.w3.org/1999/xlink'

def _BuildServiceMeta(wfs,r):
	"""Method to build the metadata etc. of the service itself
	"""

	t = etree.fromstring(r.content)	
	# General Keywords
	wfs.Keywords = [item.text for item in t.findall(_ElementKey(OWS_NAMESPACE, "ServiceIdentification/Keywords/Keyword"))]
	# Some service meta like allowed wfs versions etc
	for elem in t.findall(_ElementKey(OWS_NAMESPACE,"OperationsMetadata/Operation")):
		if elem.attrib["name"] == "GetCapabilities":
			wfs.GetCapabilitiesMeta = GetCapabilitiesMeta(elem)
		elif elem.attrib["name"] == "GetFeature":
			wfs.GetFeatureMeta = GetFeatureMeta(elem)
	# Featuretypes (Layers) and Featuretype Meta
	wfs.FeatureTypeMeta = {}
	for elem in t.findall(_ElementKey(WFS_NAMESPACE, "FeatureTypeList/FeatureType")):
		tnm = FeatureTypeMeta(elem)
		wfs.FeatureTypeMeta[tnm.FeatureType] = tnm
	wfs.FeatureTypes = tuple(
		wfs.FeatureTypeMeta.keys()
		)
	# Service contraints
	wfs.Constraints = {}
	for elem in t.findall(_ElementKey(OWS_NAMESPACE, "OperationsMetadata/Constraint")):
		dv = elem.find(_ElementKey(OWS_NAMESPACE, "DefaultValue"))
		if dv is not None:
			wfs.Constraints[elem.attrib["name"]] = dv.text
		else:
			try:
				av = elem.findall(_ElementKey(OWS_NAMESPACE, "AllowedValues/Value"))
				wfs.Constraints[elem.attrib["name"]] = [
				v.text for v in av
				]
			except Exception:
				wfs.Constraints[elem.attrib["name"]] = None
	t = None

def _BuildContentMeta(obj,elem):
	"""Method to build the content metadata
	"""

	name = elem.attrib["name"]
	# Links in the Operation content meta
	obj.RequestMethods = {}
	for e in elem.findall(_ElementKey(OWS_NAMESPACE, "DCP/HTTP/*")):
		key = e.tag.replace(f"{{{OWS_NAMESPACE}}}","")
		obj.RequestMethods.update({key.upper():e.attrib[_ElementKey(XLI_NAMESPACE, "href")]})
	# Parameters in the Operation content meta
	for e in elem.findall(_ElementKey(OWS_NAMESPACE, "Parameter")):
		key = e.attrib["name"]
		setattr(obj, key, 
			tuple(
			[item.text for item in e.findall(_ElementKey(OWS_NAMESPACE, "AllowedValues/Value"))]
			)
		)

def _BuildResonseMeta(reader, r, keyword):
	"""Method to build the metadata etc. of the geospatial data request
	"""

	t = etree.fromstring(r.content)
	# Generate Local NameSpace
	_GetLocalNS(t.nsmap)
	# Some identifiers
	reader.gml = r.content
	# Get the requested feature xml's
	reader.Features = []
	for elem in t.iter(_ElementKey(LOC_NAMESPACE, keyword)):
		reader.Features.append(Feature(elem))
	# Get the Layer meta data
	reader.LayerMeta = LayerMeta(t,keyword)	
	t = None

def _GetLocalNS(nsmap):
	"""Local Namespace of the GetCapabilities and GetFeature Response
	"""

	global LOC_NAMESPACE
	nb = ["w3.org","opengis.net"]
	b_list = [all([item not in master for item in nb]) for master in list(nsmap.values())]
	try:
		LOC_NAMESPACE = list(nsmap.values())[b_list.index(True)]
	except ValueError:
		LOC_NAMESPACE = ""

def _ElementKey(ns,sub):
	"""Return key in xml format
	"""

	def ns_string(ns,s):
		return f"{{{ns}}}{s}"
	subs = sub.split("/")
	return "/".join(tuple(map(ns_string,[ns]*len(subs),subs)))

def _IsType(elem):
	val = elem.text
	try:
		s = eval(val)
	except Exception:
		s = val
	return type(s)

def _IsFieldType(lst):
	if float in lst:
		type = float
	else:
		type = int
	if str in lst:
		type = str
	return type

class _PostElement(etree.ElementBase):
	def __init__(self,ns,sub):
		"""lxml.etree.Element to create post request data
		"""

		# Supercharge the ElementBase class
		super(_PostElement,self).__init__(nsmap={"ns0":ns})
		self.tag = _ElementKey(ns, sub)
		self.set("service","WFS")
		self.set("version","2.0.0")
		self._query = etree.SubElement(self,_ElementKey(GML_NAMESPACE, "Query"))

	def FeatureType(self,featuretype):
		"""Set the featuretype
		"""

		self._query.set("typenames",featuretype)

	def BBOXPost(self,bbox,crs):
		"""Set the bbox for the post request
		"""

		# Nested part
		f_elem = etree.SubElement(self._query, _ElementKey(FES_NAMESPACE, "Filter")) 
		bb_elem = etree.SubElement(f_elem, _ElementKey(FES_NAMESPACE, "BBOX"))
		c_elem = etree.SubElement(bb_elem, _ElementKey(GML_NAMESPACE, "Envelope"))
		# Filling it in
		c_elem.set("srsName",crs.GetURNCode())
		# Setting the bounding box coordinates
		ll = etree.SubElement(c_elem, _ElementKey(GML_NAMESPACE, "LowerCorner"))
		ll.text = f"{bbox[0]} {bbox[1]}"
		ur = etree.SubElement(c_elem, _ElementKey(GML_NAMESPACE, "UpperCorner"))
		ur.text = f"{bbox[2]} {bbox[3]}"

	def StartIndex(self, si):
		"""Set the starting index of the request
		"""

		self.set("startindex",str(si))

	def ToString(self):
		"""Return the data in xml format for the post request
		"""
		
		return etree.tostring(self)

class GetCapabilitiesMeta:
	def __init__(self,elem):
		_BuildContentMeta(self, elem)

	def __repr__(self):
		return super().__repr__()

class GetFeatureMeta:
	def __init__(self,elem):
		_BuildContentMeta(self, elem)

	def __repr__(self):
		return super().__repr__()

class FeatureTypeMeta:
	def __init__(self,elem):
		"""Create metadata of a featuretype

		Parameters
		----------
		elem : lxml.etree._Element
			Data corresponding to the featuretype in xml format
			parsed by lxml.etree

		Returns
		-------
		FeatureType metadata object
		"""

		# Identifiers
		self.FeatureType = elem.find(_ElementKey(WFS_NAMESPACE, "Name")).text
		self.Title = elem.find(_ElementKey(WFS_NAMESPACE, "Title")).text
		self.Abstract = elem.find(_ElementKey(WFS_NAMESPACE, "Abstract")).text
		# Bounding Box
		self.BBOX84 = None
		bbox = elem.find(_ElementKey(OWS_NAMESPACE, "WGS84BoundingBox"))
		if bbox is not None:
			try:
				ll = bbox.find(_ElementKey(OWS_NAMESPACE, "LowerCorner"))
				ur = bbox.find(_ElementKey(OWS_NAMESPACE, "UpperCorner"))
				self.BBOX84 = tuple(
					[float(d) for d in ll.text.split()] 
					+ [float(d) for d in ur.text.split()]
					)
			except Exception:
				self.BBOX84 = None
		# CRS
		self.CRS = tuple(
			[CRS(elem.find(_ElementKey(WFS_NAMESPACE, "DefaultCRS")).text)]
			+ [CRS(item.text) for item in elem.findall(_ElementKey(WFS_NAMESPACE, "OtherCRS"))]
			)
		# Output Formats
		self.OutputFormats = tuple(
			[item.text for item in elem.findall(_ElementKey(WFS_NAMESPACE, "OutputFormats/Format"))]
			)
		# Metadata URL
		self.MetaDataURLs = []
		for url in elem.findall(_ElementKey(WFS_NAMESPACE, "MetadataURL")):
			self.MetaDataURLs.append(url.attrib["{http://www.w3.org/1999/xlink}href"])

class Feature:
	def __init__(self,elem):
		"""Holds data of individual features returned by the request
		for geospatial data

		Parameters
		----------
		elem: lxml.etree._Element
			Data corresponding to the feature

		Returns
		-------
		Feature Object
		"""

		self.Fields = {}
		for e in elem.findall(_ElementKey(LOC_NAMESPACE, "*")):
			if e.text and e.text.strip():
				self.Fields[e.tag.replace(f"{{{LOC_NAMESPACE}}}","")] = e.text
			if e.tag.replace(f"{{{LOC_NAMESPACE}}}","").lower() \
			in ("geom","geometry","geometrie","shape"):
				self.Geometry = etree.tostring(e[0])

	def __repr__(self):
		return super().__repr__()

class LayerMeta:
	def __init__(self,t,keyword):
		"""Metadata for a shapefile layer based on gml data

		Parameters
		----------
		t: lxml.etree._Element
			gml data parsed by lxml.etree
		keyword: str
			string associated with feature dependent values
		"""

		# Headers
		self.FieldHeaders = set(
			(item.tag.replace(f"{{{LOC_NAMESPACE}}}","") 
				for item in t.iter(_ElementKey(LOC_NAMESPACE, "*")) 
				if item.text and not item.text.strip() == "")
			)
		try:
			self.FieldHeaders.remove(keyword)
		except KeyError:
			pass
		finally:
			# self.FieldHeaders = list(self.FieldHeaders)
			pass
		# Field types 
		self.FieldTypes = {}
		for header in self.FieldHeaders:
			type_list = tuple(
				map(_IsType,t.iter(_ElementKey(LOC_NAMESPACE,header)))
				)
			self.FieldTypes[header] = _IsFieldType(type_list)
		type_list = None
		# Create Header link table (max len 10 for shapefile attribute table headers)
		self.LinkTable = {}
		count = dict(zip(
			[item[0:10] for item in self.FieldHeaders],
			[0]*len(self.FieldHeaders)
			))
		for item in self.FieldHeaders:
			ab = item[0:10]
			if count[ab] >= 1:
				n = ab[:-1] + f"{count[ab]}"
			else:
				n = ab
			self.LinkTable[item] = n
			count[ab] += 1

	def __repr__(self):
		return super().__repr__()

	def __eq__(self,other):
		if isinstance(self, other.__class__):
			return sorted(self.FieldHeaders) == sorted(other.FieldHeaders)
		else:
			return False

	def __or__(self,other):
		if isinstance(self, other.__class__):
			pass
		else:
			raise TypeError(f"unsupported operand type(s) for |: '{self.__class__}' and '{other.__class__}'")

	def __ior__(self,other):
		if isinstance(self, other.__class__):
			self.FieldHeaders |= other.FieldHeaders 
			dd = defaultdict(list)
			for d in (self.FieldTypes,other.FieldTypes):
				for k,v in d.items():
					dd[k].append(v)
			for k,v in dd.items():
				self.FieldTypes.update({k:_IsFieldType(v)})
			dd = None
			self.LinkTable |= other.LinkTable
			return self
		else:
			raise TypeError(f"unsupported operand type(s) for |=: '{self.__class__}' and '{other.__class__}'")
		


from wfs20.request import parse_qsl, GetResponse 
from wfs20.util import _BuildResonseMeta

def _ServiceReader(
	url: str,
	timeout: int,
) -> 'requests.models.Response':
	"""Method to return response data for WFS service url
	"""

	r = GetResponse(url,timeout=timeout)
	return r

class DataReader:
	def __init__(
		self,
		url: str,
		keyword: str,
		method: str="GET",
		data: str=None,
	):
		"""Response reader of a geospatial data request

		Parameters
		----------
		url : str
			request url for geospatial data
		keyword : str
			Designation of the requested layer
		method : str
			Request method, either 'GET' or 'POST'
		data : str
			Params in xml format

		Returns
		-------
		DataReader
		"""

		# General stuff
		self.URL = url
		self.Keyword = keyword
		self.RequestMethod = method
		self.RequestData = data

		# substance
		_BuildResonseMeta(self, GetResponse(self.URL, timeout=30, method=method, data=data), self.Keyword)

	def __repr__(self):
		return super().__repr__()

	def __iadd__(self, other):
		if isinstance(self, other.__class__):
			self.Features += other.Features
			self.LayerMeta |= other.LayerMeta
			return self
		else:
			raise TypeError(f"unsupported operand type(s) for +=: '{self.__class__}' and '{other.__class__}'")
		


from wfs20.crs import CRS
from wfs20.error import WFSInternalError
from wfs20.io import _WriteGeometries
from wfs20.reader import _ServiceReader, DataReader
from wfs20.request import _ServiceURL, CreateGetRequest
from wfs20.util import _BuildServiceMeta

import sys

class WebFeatureService:
	def __init__(
		self,
		url: str,
		version: str="2.0.0"
	) -> 'WebFeatureService':
		"""WebFeatureService

		Parameters
		----------

		url : str
			Service url
		version : str
			WebFeatureService version
		
		Returns
		-------
		wfs20.WebFeatureService
			Object containing the metadata of the service
		"""

		self.url = url
		self.version = version
		self.ServiceURL = _ServiceURL(self.url,version)

		# Substance
		_BuildServiceMeta(self, _ServiceReader(self.ServiceURL,timeout=30))

		# Declarations
		self.DataReader = None

	def __repr__(self):
		return f"<wfs20.WebFeatureService object ({self.url})>"

	def RequestData(
		self,
		featuretype: str,
		bbox: tuple,
		epsg: int,
		):
		"""Request spatial data from the WebFeatureService

		Parameters
		----------
		featuretype : str
			Layer to be requested, mostly in the format of 'xxx:xxx'
		bbox : tuple
			Bounding box wherein the spatial data lies that is requested,
			e.g. (x1,y1,x2,y2)
		epsg : int
			The projection code of the requested data and the bounding box 
			according to EPSG, e.g. 4326 (WGS84)

		Returns
		-------
		wfs20.reader.DataReader
			Contains the requested data
		"""

		if featuretype not in self.FeatureTypes:
			raise WFSInternalError(
				"Request Error", 
				f"<{featuretype}> not in list of available featuretypes (see <class>.FeatureTypes)"
				)
		crs = CRS.from_epsg(epsg)
		if not crs in self.FeatureTypeMeta[featuretype].CRS:
			raise WFSInternalError(
				"Request Error", 
				f"<{epsg}> not in list of available projections (see <class>.FeatureTypeMeta[<id>].CRS)"
				)
		url = CreateGetRequest(
				self.url,
				self.version,
				featuretype,
				bbox,
				crs
				)
		keyword = self.FeatureTypeMeta[featuretype].Title
		self.DataReader = DataReader(url,keyword)
		return self.DataReader

	def ToFile(
		self,
		out: str,
		driver: str="GeoJSON",
		):
		"""Write geospatial data held in reader to file

		Parameters
		----------
		out : str
			path of where the file should be written to
		driver : str
			ogr driver for writing the geometries
			E.g. 'GeoJSON' if one wants to write the data
			in the geojson format
		"""
		
		if self.DataReader == None or not self.DataReader.Features:
			raise WFSInternalError("Writing to file","No features collected from WebFeatureService")
		_WriteGeometries(self.DataReader,driver,out)
		